import time
import sys
from peucr_core.loaders import ConfigLoader, SpecLoader, PluginLoader, ValidatorLoader
from peucr_core.exceptions import InvalidDefinitionException


class TestFramework:
    def __init__(self, args):
        self.separator = "**************************************************"
        self.retryInterval = 0.2
        self.config = ConfigLoader(args).apply()
        self.specs = SpecLoader(self.config).apply()
        self.plugins = PluginLoader(self.config).apply()
        self.validators = ValidatorLoader(self.config).apply()



    def exec_actions(self, actions, data):
        if data["error"] or actions is None or len(actions) == 0:
            return data

        for action in actions:
            success = False
            startTime = time.time()

            while not success and time.time() - startTime < 2:
                try:
                    r = self.plugins.apply(action)
                    success = r["success"]

                except Exception as e:
                    data["error"] = True
                    data["msg"] = "Action failed.", e

                if success or data["error"]:
                    break

                time.sleep(self.retryInterval)

        return data



    def exec_validation(self, validation, data):
        if data["error"]:
            return data

        time.sleep(validation.get("wait", 0))

        attempts = min(5, validation.get("duration", self.retryInterval)) / self.retryInterval
        counter = 0
        result = {"success": False}

        while not result["success"] and counter < attempts:
            try:
                response = self.plugins.apply(validation)
                result = self.validators.apply(validation.get("expectation"), response)

            except InvalidDefinitionException as e:
                data["msg"] = e
                break

            except Exception as e:
                data["msg"] = "Error:", e

            if result["success"]:
                break

            time.sleep(self.retryInterval)
            counter += 1

        if not result["success"]:
            data["msg"] = "Failure.", result.get("msg", "")
            data["error"] = True

        return data



    def exec_validations(self, validations, data):
        if data["error"]:
            return data

        if not validations or not isinstance(validations, list) or len(validations) == 0:
            data["error"] = True
            data["msg"] = "No validation specified in test. Aborting."
            return data

        for validation in validations:
            data = self.exec_validation(validation, data)

        return data



    def exec_test(self, spec):
        data = {"name": spec.get("name", "UNNAMED"), "start": time.time(), "error": False, "type": "test"}

        data = self.exec_actions(spec.get("context"), data)
        data = self.exec_actions(spec.get("actions"), data)
        data = self.exec_validations(spec.get("validation"), data)

        data["end"] = time.time()

        self.print_report(data)
        return data



    def exec_test_suite(self, specs):
        results = [self.exec_test(spec) for spec in specs]

        failures = len([i for i in filter(lambda x: x["error"], results)])

        print(self.separator)

        if failures > 0:
            print(len(specs), "tests run", failures, "failures")
            sys.exit(1)        

        print(len(specs), "tests run. No failures.")



    def exec_preconditions(self, validations):
        if validations is None or validations.get("validation") is None:
            return

        for validation in validations["validation"]:
            data = {"name": validation.get("name", "UNNAMED"), "type": "precondition", "start": time.time(), "error": False}

            data = self.exec_validation(validation, data)

            if data["error"]:
                data["msg"] = "Precondition validation failed. Test will be aborted"

            data["end"] = time.time()

            self.print_report(data)

            if data["error"]:
                sys.exit(1)

        print(self.separator)


    
    def print_report(self, data):
        action = "Executed" if data["type"] == "test" else "Verifying"

        print("{:s} {:s} - {:.3f}s".format(action, data["name"], data["end"]-data["start"]))

        if data["error"] and data["msg"]:
            print(data["msg"])



    def exec(self):
        self.exec_preconditions(self.specs.get("preconditions"))
        self.exec_test_suite(self.specs.get("execution"))



import boto3
import json
import time

from peucr_core.plugin import TestPlugin

class SqsReceive(TestPlugin):

    def __init__(self, config):
        self.labels = ["SQS-RECEIVE"]

        self.config = config


    def apply(self, options = {}):
        self.client = boto3.client('sqs') if "region" not in options else boto3.client('sqs', region_name=options["region"])
        
        if "sqsReceiveUrl" not in self.config:
            raise Exception("sqsReceiveUrl required in config")

        messageLimit = options["messageLimit"] if options.get("messageLimit") else 10
        duration = options["duration"] if options.get("duration") else 10
        filter = options["filter"] if options.get("filter") else {}

        messages, receipts = self.getMessages(messageLimit, filter, duration)
        self.deleteMessages(receipts)

        success = len(messages) == messageLimit
        msg = None if success else "{:d} messages received, but {:d} were expected".format(len(messages), messageLimit)
        
        return {"success": success, "msg": msg, "json": messages}


    def getMessages(self, limit, filter, duration):
        messages = {}
        start_time = time.time()

        while len(messages) < limit and time.time() < start_time + duration:
            try:
                response = self.client.receive_message(
                        QueueUrl = self.config["sqsReceiveUrl"],
                        MaxNumberOfMessages = limit,
                        WaitTimeSeconds = 2,
                        VisibilityTimeout = 0
                )

                for m in response.get("Messages", []):
                    body = json.loads(m["Body"])

                    if self.isValid(body, filter):
                        messages[m["MessageId"]] = (body, m["ReceiptHandle"])

            except Exception as e:
                pass

        return (list(m[0] for m in messages.values()), list(m[1] for m in messages.values()))


    def isValid(self, message, filter):
        for f in filter.keys():
            if f not in message:
                return False

            if message[f] != filter[f]:
                return False

        return True


    def deleteMessages(self, messages):
        for m in messages:
            self.client.delete_message(
                    QueueUrl = self.config["sqsReceiveUrl"],
                    ReceiptHandle = m
            )



from peucr_core.validators.default import DefaultValidator
from peucr_core.validators.status import StatusValidator
from peucr_core.validators.json import JsonValidator
from peucr_core.validators.oneof import OneOfValidator
from peucr_core.exceptions import InvalidDefinitionException


class ValidatorSuite:
    def __init__(self, custom):
        self.default = [StatusValidator(), JsonValidator(), OneOfValidator()]
        self.custom = custom

    def apply(self, expectation, result):
        if expectation is not None and "type" not in expectation:
            raise InvalidDefinitionException("expectation should have \"type\" defined")

        validator = self.getValidator(expectation.get("type") if expectation is not None else None)

        return validator.apply(expectation, result, self)
        

    def getValidator(self, type):
        if type is None:
            return DefaultValidator()

        validators = [v for v in self.custom if v.executes(type)]
        if validators:
            return validators[0]

        validators = [v for v in self.default if v.executes(type)]
        if validators:
            return validators[0]

        return DefaultValidator()



from peucr_core.plugins.http import HttpPlugin
from peucr_core.plugins.sqs_purge import SqsPurge
from peucr_core.plugins.sqs_receive import SqsReceive
from peucr_core.plugins.sqs_send import SqsSend
from peucr_core.exceptions import InvalidDefinitionException


class PluginSuite:
    def __init__(self, custom, config):
        self.default = [HttpPlugin(config), SqsPurge(config), SqsReceive(config), SqsSend(config)]
        self.custom = custom

    def apply(self, action):
        if action is None or (action is not None and "target" not in action):
            raise InvalidDefinitionException("action should have \"target\" defined")

        plugin = self.getPlugin(action["target"])

        return plugin.apply(action.get("options", {}))
        

    def getPlugin(self, name):
        executablePlugins = [p for p in self.custom if p.executes(name)]
        if len(executablePlugins) > 0:
            return executablePlugins[0]

        executablePlugins = [p for p in self.default if p.executes(name)]
        if len(executablePlugins) > 0:
            return executablePlugins[0]

        raise Exception("No plugin was found for " + name + ".")



import requests
from peucr_core.validator import TestValidator
from peucr_core.exceptions import InvalidDefinitionException

class OneOfValidator(TestValidator):

    def __init__(self):
        self.labels = ["ONEOF"]


    def apply(self, expectation, response, suite):
        items = expectation.get("items")
        if items is None or not isinstance(items, list) or len(items) == 0:
            raise InvalidDefinitionException("ONEOF expectation requires no empty \"items\"")

        errors = []
        for item in items:
            result = suite.apply(item, response)

            if result["success"]:
                return result
            else:
                errors.append(result["msg"])

        return {"success": False, "msg": ". ".join(errors)}



import boto3

from peucr_core.plugin import TestPlugin

class SqsSend(TestPlugin):

    def __init__(self, config):
        self.labels = ["SQS-SEND"]

        self.config = config


    def apply(self, options = {}):
        client = boto3.client('sqs') if "region" not in options else boto3.client('sqs', region_name=options["region"])

        if "sqsSendUrl" not in self.config:
            raise Exception("sqsSendUrl required in config")

        if options.get("body"):
            message = options["body"]
        else:
            raise Exception("SQS-SEND requires body in options.")

        msg = None 
        success = False

        try:
            response = client.send_message(
                QueueUrl=self.config["sqsSendUrl"],
                MessageBody=message
            )
            success = True

        except Exception as e:
            msg = e

        return {"success": success, "msg": msg}



import requests
from peucr_core.validator import TestValidator

class StatusValidator(TestValidator):

    def __init__(self):
        self.labels = ["STATUS", "STATUS-CODE"]


    def apply(self, expectation, response, suite):
        if expectation.get("value") is not None:
            success = expectation["value"] == response["status-code"]
            error = "Status code expected to be {} but was {}".format(expectation["value"], response["status-code"])
        else:
            success = response["status-code"] >= 200 and response["status-code"] < 300
            error = "Status code expected to be successful but was {}".format(response["status-code"])

        return {"success": success, "msg": error}



import boto3
import json
import time

from peucr_core.plugin import TestPlugin

class SqsPurge(TestPlugin):

    def __init__(self, config):
        self.labels = ["SQS-PURGE"]

        self.config = config


    def apply(self, options = {}):
        client = boto3.client('sqs') if "region" not in options else boto3.client('sqs', region_name=options["region"])

        if "url" not in options:
            raise Exception("url required in options")

        try:
            client.purge_queue(
                    QueueUrl = self.configure(options["url"])
            )

            return {"success": True}

        except Exception as e:
            return {"success": False, "msg": e}
